from http import HTTPStatus
from typing import List, Optional
from uuid import UUID

import requests

from user_sdk.domain import (
    Session,
    CredentialType,
    UserProfile,
    User,
    UserRequest,
    UserDevice,
)
from user_sdk.error import (
    InvalidOTPError,
    UserAuthenticationError,
    NoSuchUser,
    ProfileAlreadyExists,
    ProfileCreationError,
    ProfileUpdateError,
    NoSuchProfile,
    UserAlreadyExists,
    UserCreationFailed,
    OTPSendFailure,
    NoRoleFound,
    InvalidRoleRequest,
    SessionNotFound,
    TypeNotSupported,
    TooManyRequests,
)
from user_sdk.log import log
from user_sdk.domain import CountryCode


class UserService:
    def __init__(self, authentication_url, profile_url, authorization_url=None):
        self._authentication_url = authentication_url
        self._profile_url = profile_url
        self._authorization_url = authorization_url

    def login_with_username(self, username: str, password: str) -> Session:
        return self._login(
            cred_type=CredentialType.USERNAME, identity=username, password=password
        )

    def login_with_mobile(
        self,
        phone_number: str,
        otp: str,
        region: CountryCode = CountryCode.INDIA,
        expiry_seconds: Optional[int] = None,
    ) -> Session:
        return self._login(
            cred_type=CredentialType.MOBILE,
            identity=phone_number,
            otp=otp,
            region=region,
            expiry_seconds=expiry_seconds,
        )

    def login_with_oauth(self, id_token: str) -> Session:
        return self._login(cred_type=CredentialType.OAUTH, identity=id_token)

    def _login(
        self,
        cred_type,
        identity,
        password=None,
        otp=None,
        region=CountryCode.INDIA,
        expiry_seconds: Optional[int] = None,
    ) -> Session:
        body = {
            "identity": identity,
            "credential_type": str(cred_type),
            "region": str(region),
        }
        if cred_type == CredentialType.USERNAME:
            body["password"] = password
        elif cred_type == CredentialType.MOBILE:
            body["otp"] = otp
            if expiry_seconds:
                body["expiry_seconds"] = expiry_seconds

        request_url = f"{self._authentication_url}/api/v1/sign_in"
        response = requests.post(request_url, json=body)
        log(
            message="login",
            request_url=request_url,
            request_body=body,
            status_code=response.status_code,
            response=response.text,
        )
        if response.status_code == HTTPStatus.CREATED:
            return Session.from_dict(response.json()["data"])
        if response.status_code == HTTPStatus.BAD_REQUEST:
            error_type = response.json()["error"]["type"]
            if error_type == "INVALID_PASSWORD" or error_type == "INVALID_OTP":
                raise InvalidOTPError
        if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            raise TooManyRequests

        raise UserAuthenticationError(response.json()["error"])

    def get_user_from_session(self, session_id: str) -> Session:
        response = requests.get(
            f"{self._authentication_url}/api/v1/sessions/{session_id}"
        )
        if response.status_code == HTTPStatus.OK:
            return Session.from_dict(response.json()["data"])
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchUser

        self._raise_response_error(response)

    def get_user_from_active_session(self, session_id: str) -> Session:
        response = requests.get(
            f"{self._authentication_url}/api/v1/sessions/{session_id}/active"
        )
        if response.status_code == HTTPStatus.OK:
            return Session.from_dict(response.json()["data"])
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchUser

        self._raise_response_error(response)

    def clear_all_sessions(self, phone_number: str):
        response = requests.delete(
            f"{self._authentication_url}/api/v1/sessions/clear_all_sessions",
            json={"phone_number": phone_number},
        )
        if response.status_code == HTTPStatus.OK:
            return response.json()["data"]
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchUser

        self._raise_response_error(response)

    def attach_device(self, user_device: UserDevice):
        request_url = f"{self._profile_url}/api/v1/devices"
        request_json = user_device.to_dict()
        response = requests.post(request_url, json=request_json)
        log(
            message="attach_device",
            request_url=request_url,
            status_code=response.status_code,
            response=response.text,
            request_body=request_json,
        )

        if response.status_code == HTTPStatus.CREATED:
            return UserDevice.from_dict(response.json()["data"])

        if response.status_code == HTTPStatus.CONFLICT:
            return

        self._raise_response_error(response)

    def get_all_device_entries_for_device_id(self, device_id: str) -> List[UserDevice]:
        request_url = f"{self._profile_url}/api/v1/devices/{device_id}"
        response = requests.get(request_url)
        if response.status_code == HTTPStatus.OK:
            return [
                UserDevice.from_dict(data_dict) for data_dict in response.json()["data"]
            ]
        self._raise_response_error(response)

    def create_profile(self, profile: UserProfile) -> UserProfile:
        request_url = f"{self._profile_url}/api/v1/profiles"
        response = requests.put(request_url, json={"profile": profile.to_dict()})
        log(
            message="create_profile",
            request_url=request_url,
            request_body=profile.to_dict(),
            status_code=response.status_code,
            response=response.text,
        )

        if response.status_code == HTTPStatus.CREATED:
            return UserProfile.from_dict(response.json()["data"])
        if response.status_code == HTTPStatus.CONFLICT:
            raise ProfileAlreadyExists()
        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise ProfileCreationError(response.json().get("error"))

        self._raise_response_error(response)

    def bulk_upsert_profiles(self, profiles: List[UserProfile]) -> List[UserProfile]:
        request_url = f"{self._profile_url}/api/v1/profiles/bulk"
        request_json = {"profiles": [profile.to_dict() for profile in profiles]}
        response = requests.put(request_url, json=request_json)
        log(
            message="bulk_upsert_profiles",
            request_url=request_url,
            request_body=request_json,
            status_code=response.status_code,
            response=response.text,
        )

        if response.status_code == HTTPStatus.OK:
            return [
                UserProfile.from_dict(profile) for profile in response.json()["data"]
            ]
        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise ProfileCreationError(response.json().get("error"))

        self._raise_response_error(response)

    def update_profile(self, user_id: UUID, profile: dict) -> UserProfile:
        request_url = f"{self._profile_url}/api/v1/new/profiles/{user_id}"
        response = requests.patch(url=request_url, json=profile)
        log(
            message="update_profile",
            request_url=request_url,
            status_code=response.status_code,
            response=response.text,
            request_body=profile,
        )
        if response.status_code == HTTPStatus.OK:
            return UserProfile.from_dict(response.json().get("data"))
        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise ProfileUpdateError(response.json().get("error"))
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchProfile
        if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            raise TooManyRequests

        self._raise_response_error(response)

    def get_user_profile(self, id: UUID) -> UserProfile:
        request_url = f"{self._profile_url}/api/v1/profiles/{id}"
        response = requests.get(request_url)
        log(
            message="get_user_profile",
            request_url=request_url,
            status_code=response.status_code,
            response=response.text,
        )
        if response.status_code == HTTPStatus.OK:
            return UserProfile.from_dict(response.json().get("data"))
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchProfile

        self._raise_response_error(response)

    def get_user_profiles(self, ids: List[UUID]) -> List[UserProfile]:
        request_url = f"{self._profile_url}/api/v1/profiles/by_user_ids"
        response = requests.get(request_url, params={"ids": ",".join(map(str, ids))})
        log(
            message="get_user_profile",
            request_url=request_url,
            status_code=response.status_code,
            response=response.text,
        )
        if response.status_code == HTTPStatus.OK:
            return [UserProfile.from_dict(up) for up in response.json()["data"]]

        self._raise_response_error(response)

    def create_user(self, user_request: UserRequest):
        request_url = f"{self._authentication_url}/api/v1/users"

        response = requests.post(request_url, json=user_request.to_dict())
        log(
            message="create_user",
            request_url=request_url,
            status_code=response.status_code,
            response=response.text,
            request_body=user_request.to_dict(),
        )

        if response.status_code == HTTPStatus.CREATED:
            return User.from_dict(response.json().get("data"))
        if response.status_code == HTTPStatus.CONFLICT:
            raise UserAlreadyExists
        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise UserCreationFailed(response.json().get("error"))
        else:
            raise OTPSendFailure(response.json().get("error"))

    def bulk_create_users(self, user_requests: List[UserRequest]):
        request_json = {"users": [ur.to_dict() for ur in user_requests]}
        request_url = f"{self._authentication_url}/api/v1/users/bulk"
        response = requests.post(request_url, json=request_json)
        log(
            message="bulk_create_users",
            request_url=request_url,
            status_code=response.status_code,
            response=response.text,
            request_body=request_json,
        )

        if response.status_code == HTTPStatus.OK:
            return [User.from_dict(user) for user in response.json().get("data")]
        if response.status_code == HTTPStatus.CONFLICT:
            raise UserAlreadyExists
        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise UserCreationFailed(response.json().get("error"))
        else:
            raise OTPSendFailure(response.json().get("error"))

    def generate_otp(
        self, credential_type: str, identity, region: str = CountryCode.INDIA
    ):
        cred_type = CredentialType(credential_type)
        body = {
            "credential_type": cred_type.name,
            "identity": identity,
            "region": str(region),
        }
        request_url = f"{self._authentication_url}/api/v1/sessions/otp"
        response = requests.post(request_url, json=body)
        log(
            message="generate_otp",
            request_url=request_url,
            status_code=response.status_code,
            response=response.text,
            request_body=body,
        )
        if response.status_code != HTTPStatus.ACCEPTED:
            raise OTPSendFailure

    def validate_otp(
        self,
        credential_type: str,
        identity: str,
        otp: str,
        region: CountryCode = CountryCode.INDIA,
    ) -> bool:
        if credential_type != CredentialType.MOBILE.value:
            raise TypeNotSupported

        cred_type = CredentialType(credential_type)
        params = {
            "credential_type": cred_type.name,
            "identity": identity,
            "otp": otp,
            "region": str(region),
        }
        request_url = f"{self._authentication_url}/api/v1/sessions/otp/validate"
        response = requests.get(request_url, params=params)
        if response.status_code == HTTPStatus.OK:
            data = response.json()["data"]
            return data["valid"]
        if response.status_code == HTTPStatus.TOO_MANY_REQUESTS:
            raise TooManyRequests

        self._raise_response_error(response)

    def get_otp(
        self, credential_type: str, identity: str, region=CountryCode.INDIA
    ) -> str:
        if credential_type != CredentialType.MOBILE.value:
            raise TypeNotSupported

        cred_type = CredentialType(credential_type)
        params = {
            "credential_type": cred_type.name,
            "identity": identity,
            "region": region,
        }
        request_url = f"{self._authentication_url}/api/v1/sessions/otp"
        response = requests.get(request_url, params=params)
        if response.status_code == HTTPStatus.OK:
            data = response.json()["data"]
            return data["otp"]

        self._raise_response_error(response)

    def get_by_email(self, email: str) -> User:
        response = requests.get(
            f"{self._authentication_url}/api/v1/users/by_identity/%s" % email
        )
        if response.status_code == HTTPStatus.OK:
            return User.from_dict(response.json().get("data"))
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchUser

        self._raise_response_error(response)

    def get_by_mobile_number(self, mobile_number: str) -> User:
        response = requests.get(
            f"{self._authentication_url}/api/v1/users/by_identity/%s" % mobile_number
        )
        if response.status_code == HTTPStatus.OK:
            return User.from_dict(response.json().get("data"))
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchUser

        self._raise_response_error(response)

    def get_by_identities(self, identities: List[str]) -> List[User]:
        response = requests.get(
            f"{self._authentication_url}/api/v1/users/by_identities",
            params={"identities": ",".join(identities)},
        )
        if response.status_code == HTTPStatus.OK:
            return [User.from_dict(u) for u in response.json().get("data")]

        self._raise_response_error(response)

    def get_by_ids(self, ids: List[str]) -> List[User]:
        response = requests.get(
            f"{self._authentication_url}/api/v1/users/by_user_ids",
            params={"ids": ",".join(ids)},
        )
        if response.status_code == HTTPStatus.OK:
            return [User.from_dict(u) for u in response.json().get("data")]

        self._raise_response_error(response)

    def get_user(self, user_id: UUID) -> User:
        response = requests.get(f"{self._authentication_url}/api/v1/users/{user_id}")
        if response.status_code == HTTPStatus.OK:
            return User.from_dict(response.json()["data"])
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchUser

        self._raise_response_error(response)

    def get_user_role(self, email: str, panel: str) -> str:
        response = requests.get(
            f"{self._authorization_url}/api/v1/role",
            params=dict(email=email, panel=panel),
        )

        if response.status_code == HTTPStatus.OK:
            return response.json()["data"]["role"]

        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoRoleFound("No role found for this email on panel")

        if response.status_code == HTTPStatus.BAD_REQUEST:
            raise InvalidRoleRequest("Either Panel name or email format is invalid")

        self._raise_response_error(response)

    def delete_session(self, session_id: str):
        response = requests.delete(
            f"{self._authentication_url}/api/v1/sessions/{session_id}"
        )
        if response.status_code == HTTPStatus.OK:
            return response.json()["data"]
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise SessionNotFound

        self._raise_response_error(response)

    def fetch_sessions_by_phone_number(self, phone_number: str) -> List[Session]:
        response = requests.get(
            f"{self._authentication_url}/api/v1/sessions/by_phone_number/{phone_number}"
        )
        if response.status_code == HTTPStatus.OK:
            sessions = response.json()["data"]
            return [Session.from_dict(s_dict) for s_dict in sessions]
        if response.status_code == HTTPStatus.NOT_FOUND:
            raise NoSuchUser

        self._raise_response_error(response)

    def _raise_response_error(self, response):
        raise RuntimeError(
            f"""
            Invalid response.

            Status code: {response.status_code}
            Text:

            {response.text}
        """
        )



from unittest.mock import patch

import responses
import pytest

from user_sdk.openid_connect import (
    OpenIDConnectAuth,
    OIDCAuthError,
    InsufficientPermissionError,
)


def _jw_key_set():
    return {
        "keys": [
            {
                "alg": "RS256",
                "kty": "RSA",
                "use": "sig",
                "n": "60U6mBmsm6kHKqv9Q8iYv_QTfyhjL0LpJQOVEbZsWdI9nxJzh2Clmro6tv_zvZsaLhl4BJ392gdATGrcYqnAooNa-6pDG9Z-WrhEkTHna-bcqB_LaROpdCW8InJlMvWg4CdPE0lfDHGjhIA8AKNk0Hh7-zPkoNeqrizgCZR_COSW-I3jzWF4m5UumGQqGXxjUcGipyZVmy8qyrQ8-xaTyHxFxgqnOsypkPaDZzjuSjCaXCpqDlC1-uW3qeB-SRD0XFqp_4lNm65HuvGeIzlVMgAunpTqJbm4seknnwbWvedotm4L8XxOnZIFI0uTQ5PqSGsx8N5TA-butSkKLY5TDQ",
                "e": "AQAB",
                "kid": "N_68xPq0IVA9TJkB9fB4j",
            },
            {
                "alg": "RS256",
                "kty": "RSA",
                "use": "sig",
                "n": "uzG86qsHYs0PPIoFPe1IIIYmh3fBbgWPD2xpxiPkOEajoohg8_YkAYbFp2gZhFosO2L88le5DpmDSsjwsWUXTcr0cYnAnOPMoZUE--rwGezHyvt4Tq4k3DQf3szSgbRm4DrUBLqr368eJ9pPkIyzaTxQmLjgp2_WZ2vTuhRiU0UAMiwLkrVkcO_K42TYHeci9en5uYajAUAxLwFnUlBQYEEU7p8NQVeCHMJ0-LWiPIJT3SAz6uyglYJjNV3rsJ2wDJjBmT44l6otHmLDVM53mTtMTf-hBB1Gl9QbSPL3_r4Xsul-VvscJcCxUNlluHO-ehyAak4DgwFdFy7gqdeI4w",
                "e": "AQAB",
                "kid": "S0T8Tr7PjB4FIHRCsXptG",
            },
        ]
    }


def _user_info(email):
    return {
        "sub": "auth0|5f649a4215ab3a00772db6a6",
        "nickname": "Rodya",
        "name": "Rodion Romanovich Raskolnikov",
        "picture": "https://www.some.picture",
        "updated_at": "2020-09-18T11:30:11.171Z",
        "email": email,
        "email_verified": False,
    }


@patch("jose.jwt.get_unverified_claims")
@patch("jose.jwt.get_unverified_header")
@patch("jose.jwt.decode")
@responses.activate
def test_authorize(mock_jwt_decode, mock_get_headers, mock_get_claims):
    jw_key_set = _jw_key_set()
    email = "e@m.ail"
    responses.add("GET", "https://issuer.url/.well-known/jwks.json", json=jw_key_set)
    responses.add("GET", "https://issuer.url/userinfo", json=_user_info(email))
    mock_get_headers.return_value = {"kid": jw_key_set["keys"][0]["kid"]}
    mock_get_claims.return_value = {"permissions": ["read:abc"]}

    sso = OpenIDConnectAuth(
        issuer_url="https://issuer.url", audience="test-audience", algorithms=["RS256"],
    )
    assert sso.authorize(access_token="token", required_permission="read:abc") == email
    key_set = jw_key_set["keys"][0].copy()
    key_set.pop("alg")
    mock_jwt_decode.assert_called_with(
        "token",
        key_set,
        algorithms=["RS256"],
        audience="test-audience",
        issuer="https://issuer.url/",
    )
    mock_get_headers.assert_called_with("token")
    mock_get_claims.assert_called_with("token")
    responses.assert_call_count("https://issuer.url/.well-known/jwks.json", 1)
    responses.assert_call_count("https://issuer.url/userinfo", 1)


@patch("jose.jwt.get_unverified_claims")
@patch("jose.jwt.get_unverified_header")
@patch("jose.jwt.decode")
@responses.activate
def test_authorize_with_insufficient_permission(
    mock_jwt_decode, mock_get_headers, mock_get_claims
):
    jw_key_set = _jw_key_set()
    responses.add("GET", "https://issuer.url/.well-known/jwks.json", json=jw_key_set)
    responses.add("GET", "https://issuer.url/userinfo", json=_user_info("e@m.ail"))
    mock_get_headers.return_value = {"kid": jw_key_set["keys"][0]["kid"]}
    mock_get_claims.return_value = {"permissions": ["read:abc"]}

    sso = OpenIDConnectAuth(
        issuer_url="https://issuer.url", audience="test-audience", algorithms=["RS256"],
    )

    with pytest.raises(InsufficientPermissionError):
        assert sso.authorize(access_token="token", required_permission="read:xyz")
    responses.assert_call_count("https://issuer.url/userinfo", 0)


@pytest.mark.parametrize(
    "granted_permissions, required_permission, result",
    [
        (["read:vehicles", "write:vehicles"], "write:vehicles", True),
        (["read:vehicles", "write:vehicles"], "read:vehicles", True),
        (["read:vehicles", "write:vehicles"], "read:persons", False),
        (["read:*", "write:vehicles", "read:stops"], "read:persons", True),
        (["read:vehicles", "write:*"], "write:persons", True),
        ([], "read:vehicles", False),
        (["*:persons"], "read:persons", False),
    ],
)
@patch("jose.jwt.get_unverified_claims")
def test_has_permission(
    mock_get_unverified_claims, granted_permissions, required_permission, result
):
    mock_get_unverified_claims.return_value = {"permissions": granted_permissions}
    assert OpenIDConnectAuth._has_permission("abc-token", required_permission) == result



import json
from uuid import uuid4, UUID
from http import HTTPStatus

import pytest
import responses
from user_sdk.domain import User, UserProfile, Session
from user_sdk.error import TooManyRequests
from user_sdk.user_service import UserService

user_service = UserService(
    authentication_url="http://dummy_authentication",
    profile_url="http://dummy_profile",
    authorization_url="http://dummy_authorization",
)


def test_dict_to_user():
    user = User.from_dict(user_dict)
    assert user.id == "b402e055-2884-43a1-9e33-dbfe67e62971"
    assert 1 == len(user.identities)


def test_dict_to_user_profile():
    user_profile = UserProfile.from_dict(user_profile_dict)
    assert user_profile.name == "sample_name"
    assert "sample_work_address" == user_profile.work_address.location_name
    assert "sample_home_address" == user_profile.home_address.location_name
    assert "MALE" == user_profile.gender.name
    assert "gcm-is-fcm-is-gcm" == user_profile.push_notification_id


def test_dict_to_user_profile_for_incomplete_user_profile():
    user_profile = UserProfile.from_dict(incomplete_user_profile_dict)
    assert user_profile.name is None
    assert user_profile.work_address is None
    assert "sample_home_address" == user_profile.home_address.location_name
    assert user_profile.gender is None


def test_user_profile_to_dict():
    user_profile = UserProfile.from_dict(user_profile_dict)
    user_profile_new_dict = user_profile.to_dict()

    for key in [
        "id",
        "name",
        "gender",
        "home_address",
        "work_address",
        "user_types",
        "zone_id",
        "user_specific_features",
    ]:
        assert user_profile_dict[key] == user_profile_new_dict[key]


def test_user_profile_with_no_home_work_address_to_dict():
    user_profile = UserProfile.from_dict(user_profile_dict_with_no_home_or_work_address)
    user_profile_new_dict = user_profile.to_dict()

    for key in ["id", "name", "gender"]:
        assert (
            user_profile_dict_with_no_home_or_work_address[key]
            == user_profile_new_dict[key]
        )


def test_user_profile_with_no_name_gender_to_dict():
    user_profile = UserProfile.from_dict(user_profile_dict_with_no_name_or_gender)
    user_profile_new_dict = user_profile.to_dict()

    assert user_profile_dict_with_no_name_or_gender["id"] == user_profile_new_dict["id"]


def test_dict_to_session():
    session = Session.from_dict(session_response)
    assert session.user.id == "30944af3-dfe7-45e5-a860-31ac4a18bc69"
    assert (
        session.id
        == "FEZRHOQLBALFB4QE3S3LLN52RROVR6PTC4JQT44WFFZWNC4QKT3JTFQJX2E2LAKXGCFQD4BN63TTEIADCFKDFYAI4RCJRAHL7JCWHYY"
    )


def test_add_user_type_to_user_profile():
    user_profile = UserProfile.from_dict(incomplete_user_profile_dict)
    assert user_profile.user_types == []

    user_profile.add_user_type("PREMIUM")
    assert user_profile.user_types == ["PREMIUM"]


def test_remove_user_type_to_user_profile():
    user_profile = UserProfile.from_dict(user_profile_dict)
    assert user_profile.user_types == ["PREMIUM"]

    user_profile.remove_user_type("DUMMY")
    assert user_profile.user_types == ["PREMIUM"]

    user_profile.remove_user_type("PREMIUM")
    assert user_profile.user_types == []


def test_add_user_specific_feature_to_user_profile():
    user_profile = UserProfile.from_dict(incomplete_user_profile_dict)
    assert user_profile.user_specific_features == []

    user_profile.add_user_specific_feature("AUTO_BOOKING")
    assert user_profile.user_specific_features == ["AUTO_BOOKING"]


def test_remove_user_specific_feature_from_user_profile():
    user_profile = UserProfile.from_dict(user_profile_dict)
    assert user_profile.user_specific_features == ["AUTO_BOOKING"]

    user_profile.remove_user_specific_feature("DUMMY")
    assert user_profile.user_specific_features == ["AUTO_BOOKING"]

    user_profile.remove_user_specific_feature("AUTO_BOOKING")
    assert user_profile.user_specific_features == []


@responses.activate
def test_update_profile_on_adding_user_types_and_user_specific_features():
    user_profile = UserProfile.from_dict(incomplete_user_profile_dict)
    user_id = UUID(incomplete_user_profile_dict["id"])

    def request_callback(request):
        payload = json.loads(request.body)
        assert payload["user_types"] == ["PREMIUM"]
        assert payload["user_specific_features"] == ["AUTO_BOOKING"]
        return (
            200,
            {},
            json.dumps(
                {
                    "data": {
                        "id": str(user_id),
                        "user_types": payload["user_types"],
                        "user_specific_features": payload["user_specific_features"],
                    }
                }
            ),
        )

    responses.add_callback(
        responses.PATCH,
        f"http://dummy_profile/api/v1/new/profiles/{user_id}",
        callback=request_callback,
        content_type="application/json",
    )

    user_profile.add_user_type("PREMIUM")
    user_profile.add_user_specific_feature("AUTO_BOOKING")
    user_service.update_profile(
        user_id=user_id,
        profile={
            "user_types": user_profile.user_types,
            "user_specific_features": user_profile.user_specific_features,
        },
    )


@responses.activate
def test_update_profile_returns_429():
    user_id = UUID(incomplete_user_profile_dict["id"])

    responses.add(
        responses.PATCH,
        f"http://dummy_profile/api/v1/new/profiles/{user_id}",
        content_type="application/json",
        status=429,
    )

    with pytest.raises(TooManyRequests):
        user_service.update_profile(
            user_id=user_id, profile={"push_notification_id": "fcm-is-gcm-is-fcm"}
        )


@responses.activate
def test_update_profile_on_removing_user_types_and_user_specific_features():
    user_profile = UserProfile.from_dict(user_profile_dict)
    user_id = UUID(incomplete_user_profile_dict["id"])

    def request_callback(request):
        payload = json.loads(request.body)
        assert payload["user_types"] == []
        assert payload["user_specific_features"] == []
        return (
            200,
            {},
            json.dumps(
                {
                    "data": {
                        "id": str(user_id),
                        "user_types": payload["user_types"],
                        "user_specific_features": payload["user_specific_features"],
                    }
                }
            ),
        )

    responses.add_callback(
        responses.PATCH,
        f"http://dummy_profile/api/v1/new/profiles/{user_id}",
        callback=request_callback,
        content_type="application/json",
    )

    user_profile.remove_user_type("PREMIUM")
    user_profile.remove_user_specific_feature("AUTO_BOOKING")
    user_service.update_profile(
        user_id=user_id,
        profile={
            "user_types": user_profile.user_types,
            "user_specific_features": user_profile.user_specific_features,
        },
    )


@responses.activate
def test_get_user_from_session_throws_rte_if_non_ok_response():
    session_id = str(uuid4())
    responses.add(
        responses.GET,
        f"http://dummy_authentication/api/v1/sessions/{session_id}",
        status=500,
        json=dict(key="value"),
    )
    with pytest.raises(RuntimeError) as exc_info:
        user_service.get_user_from_session(session_id)

    assert "500" in str(exc_info.value)
    assert "key" in str(exc_info.value)
    assert "value" in str(exc_info.value)


@responses.activate
def test_get_user_from_active_session_for_ok_response():
    session_id = str(uuid4())

    def request_callback(request):
        return HTTPStatus.OK, {}, json.dumps({"data": session_response})

    responses.add_callback(
        responses.GET,
        f"http://dummy_authentication/api/v1/sessions/{session_id}/active",
        callback=request_callback,
        content_type="application/json",
    )
    session = user_service.get_user_from_active_session(session_id)
    assert Session.from_dict(session_response) == session


@responses.activate
def test_get_user_from_active_session_throws_rte_if_non_ok_response():
    session_id = str(uuid4())
    responses.add(
        responses.GET,
        f"http://dummy_authentication/api/v1/sessions/{session_id}/active",
        status=500,
        json=dict(key="value"),
    )
    with pytest.raises(RuntimeError) as exc_info:
        user_service.get_user_from_active_session(session_id)

    assert "500" in str(exc_info.value)
    assert "key" in str(exc_info.value)
    assert "value" in str(exc_info.value)


user_dict = {
    "created_at": "2018-10-18T09:05:59.334113",
    "credentials": [
        {
            "created-at": "2018-10-18 09:05:59.334075",
            "id": "d417ad09-d5ae-4a72-9060-419308e6756d",
            "identity": "+917042329887",
            "identity_type": "MOBILE",
            "updated-at": "2018-10-18 09:05:59.334091",
            "verified": True,
        }
    ],
    "id": "b402e055-2884-43a1-9e33-dbfe67e62971",
    "suspended": False,
    "updated_at": "2018-10-18T09:05:59.334119",
}

user_profile_dict = {
    "created_at": "2018-10-18T13:04:55.643673",
    "gender": "MALE",
    "home_address": {
        "location": {"lat": 28.88, "lng": 77.33},
        "location_name": "sample_home_address",
    },
    "id": "f77971a6-3a7f-4475-81b6-2fe9e085f0bc",
    "name": "sample_name",
    "zone_id": "b402e055-2884-43a1-9e33-dbfe67e62971",
    "updated_at": "2018-10-18T13:04:55.643673",
    "push_notification_id": "gcm-is-fcm-is-gcm",
    "work_address": {
        "location": {"lat": 27.88, "lng": 76.33},
        "location_name": "sample_work_address",
    },
    "user_types": ["PREMIUM"],
    "user_specific_features": ["AUTO_BOOKING"],
}

user_profile_dict_with_no_home_or_work_address = {
    "id": "f77971a6-3a7f-4475-81b6-2fe9e085f0bc",
    "gender": "MALE",
    "name": "Test",
}

user_profile_dict_with_no_name_or_gender = {
    "id": "f77971a6-3a7f-4475-81b6-2fe9e085f0bc"
}

incomplete_user_profile_dict = {
    "created_at": "2018-10-18T13:04:55.643673",
    "home_address": {
        "location": {"lat": 28.88, "lng": 77.33},
        "location_name": "sample_home_address",
    },
    "id": "f77971a6-3a7f-4475-81b6-2fe9e085f0bc",
    "updated_at": "2018-10-18T13:04:55.643673",
}

session_response = {
    "id": "fd335403-83bc-4eb5-b42e-6eb6674f4537",
    "created_at": "2018-10-21T13:52:50.208448",
    "updated_at": "2018-10-21T13:52:50.208457",
    "session_id": "FEZRHOQLBALFB4QE3S3LLN52RROVR6PTC4JQT44WFFZWNC4QKT3JTFQJX2E2LAKXGCFQD4BN63TTEIADCFKDFYAI4RCJRAHL7JCWHYY",
    "user": {
        "id": "30944af3-dfe7-45e5-a860-31ac4a18bc69",
        "created_at": "2018-10-21T07:41:36.034161",
        "updated_at": "2018-10-21T07:41:36.034184",
        "suspended": False,
        "credentials": [
            {
                "id": "c28ee1f5-8632-42e4-97a3-bbd60893bb7b",
                "created-at": "2018-10-21 07:41:36.023701",
                "updated-at": "2018-10-21 07:41:36.032386",
                "identity": "+917042329887",
                "identity_type": "MOBILE",
                "verified": True,
            }
        ],
    },
    "credential": {
        "id": "c28ee1f5-8632-42e4-97a3-bbd60893bb7b",
        "created-at": "2018-10-21 07:41:36.023701",
        "updated-at": "2018-10-21 07:41:36.032386",
        "identity": "+917042329887",
        "identity_type": "MOBILE",
        "verified": True,
    },
}


@responses.activate
def test_login_raises_too_many_requests_error_if_status_code_received_from_user_auth_service_is_429():
    responses.add(
        responses.POST,
        f"http://dummy_authentication/api/v1/sign_in",
        json={},
        status=HTTPStatus.TOO_MANY_REQUESTS,
    )
    with pytest.raises(TooManyRequests):
        user_service.login_with_mobile(phone_number="8130813081", otp="1339")


@responses.activate
def test_mobile_login_with_expiry():
    expiry = 600

    def request_callback(request):
        payload = json.loads(request.body)
        assert payload["expiry_seconds"] == expiry
        return (
            HTTPStatus.CREATED,
            {},
            json.dumps({"data": {**session_response, "expiry_seconds": expiry}}),
        )

    responses.add_callback(
        responses.POST,
        f"http://dummy_authentication/api/v1/sign_in",
        callback=request_callback,
        content_type="application/json",
    )
    user_service.login_with_mobile(
        phone_number="8130813081", otp="1339", expiry_seconds=expiry
    )



from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Optional, List
from uuid import UUID

from shuttl_geo import Location
from shuttlis.serialization import serialize


class CredentialType(Enum):
    USERNAME = "USERNAME"
    MOBILE = "MOBILE"
    OAUTH = "OAUTH"

    def __str__(self):
        return self.value


class CountryCode(Enum):
    INDIA = "IN"
    THAILAND = "TH"

    def __str__(self):
        return self.value


class Gender(Enum):
    MALE = "MALE"
    FEMALE = "FEMALE"
    OTHER = "OTHER"

    def __str__(self):
        return self.value


@dataclass
class Address:
    location_name: str
    location: Location
    street_address: Optional[str] = None
    usual_departure_time: Optional[int] = None

    @classmethod
    def from_dict(cls, _dict: dict) -> "Address":
        return cls(
            location=Location(_dict["location"]["lat"], _dict["location"]["lng"]),
            location_name=_dict["location_name"],
            street_address=_dict.get("street_address"),
            usual_departure_time=_dict.get("usual_departure_time"),
        )

    def to_dict(self):
        rv = {"location_name": self.location_name, "location": serialize(self.location)}

        if self.street_address:
            rv["street_address"] = self.street_address

        if self.usual_departure_time:
            rv["usual_departure_time"] = self.usual_departure_time

        return rv


@dataclass
class UserProfile:
    user_id: UUID
    name: str
    gender: Gender
    home_address: Address
    work_address: Address
    zone_id: Optional[str] = None
    dob: Optional[datetime] = None
    email: Optional[str] = None
    img_url: Optional[str] = None
    push_notification_id: Optional[str] = None
    fb_access_token: Optional[str] = None
    user_types: List[str] = field(default_factory=list)
    user_specific_features: List[str] = field(default_factory=list)
    features: List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, _dict) -> "UserProfile":
        return UserProfile(
            user_id=_dict["id"],
            name=_dict.get("name"),
            gender=Gender(_dict["gender"]) if _dict.get("gender") else None,
            zone_id=_dict.get("zone_id"),
            home_address=Address.from_dict(_dict["home_address"])
            if _dict.get("home_address")
            else None,
            work_address=Address.from_dict(_dict["work_address"])
            if _dict.get("work_address")
            else None,
            email=_dict.get("email"),
            img_url=_dict.get("img_url"),
            push_notification_id=_dict.get("push_notification_id"),
            dob=datetime.fromisoformat(_dict["dob"]) if _dict.get("dob") else None,
            fb_access_token=_dict.get("fb_access_token"),
            user_types=_dict.get("user_types") or [],
            user_specific_features=_dict.get("user_specific_features") or [],
            features=_dict.get("features") or [],
        )

    def to_dict(self):
        rv = {"id": str(self.user_id)}
        if self.name:
            rv["name"] = self.name
        if self.gender:
            rv["gender"] = self.gender.name
        if self.home_address:
            rv["home_address"] = self.home_address.to_dict()
        if self.work_address:
            rv["work_address"] = self.work_address.to_dict()

        if self.zone_id:
            rv["zone_id"] = self.zone_id
        if self.dob:
            rv["dob"] = self.dob
        if self.email:
            rv["email"] = self.email
        if self.img_url:
            rv["img_url"] = self.img_url
        if self.push_notification_id:
            rv["push_notification_id"] = self.push_notification_id
        if self.fb_access_token:
            rv["fb_access_token"] = self.fb_access_token
        if self.user_types:
            rv["user_types"] = self.user_types
        if self.user_specific_features:
            rv["user_specific_features"] = self.user_specific_features
        return rv

    def add_user_type(self, type: str):
        if type not in self.user_types:
            self.user_types.append(type)

    def remove_user_type(self, type: str):
        if type in self.user_types:
            self.user_types.remove(type)

    def add_user_specific_feature(self, feature: str):
        if feature not in self.user_specific_features:
            self.user_specific_features.append(feature)

    def remove_user_specific_feature(self, feature: str):
        if feature in self.user_specific_features:
            self.user_specific_features.remove(feature)


@dataclass
class UserDevice:
    user_id: UUID
    device_id: str
    app_version: str = None
    platform: str = None

    @classmethod
    def from_dict(cls, _dict: dict):
        return cls(
            user_id=_dict["user_id"],
            device_id=_dict["device_id"],
            app_version=_dict.get("app_version"),
            platform=_dict.get("platform"),
        )

    def to_dict(self):
        rv = {"user_id": str(self.user_id), "device_id": self.device_id}

        if self.app_version:
            rv["app_version"] = self.app_version

        if self.platform:
            rv["platform"] = self.platform

        return rv


@dataclass
class Credential:
    id: UUID
    type: CredentialType
    identity: str
    verified: bool

    @classmethod
    def from_dict(cls, _dict: dict) -> "Credential":
        return Credential(
            id=_dict["id"],
            type=CredentialType(_dict["identity_type"]),
            identity=_dict["identity"],
            verified=_dict["verified"],
        )


@dataclass
class User:
    id: UUID
    identities: [Credential]

    @classmethod
    def from_dict(cls, _dict) -> "User":
        return cls(
            id=_dict["id"],
            identities=[Credential.from_dict(cred) for cred in _dict["credentials"]],
        )

    @property
    def primary_cred(self):
        return self.identities[0]

    @property
    def primary_identity(self):
        return self.primary_cred.identity


@dataclass
class Session:
    id: str
    user: User

    @classmethod
    def from_dict(cls, _dict) -> "Session":
        user = User.from_dict(_dict.get("user"))
        return Session(_dict.get("session_id"), user=user)


@dataclass
class UserRequest:
    credential_type: CredentialType
    identity: str
    requires_verification: bool = True
    password: str = None
    region: str = CountryCode.INDIA
    id: str = None

    def to_dict(self):
        request = {
            "credential_type": self.credential_type.value,
            "identity": self.identity,
            "requires_verification": self.requires_verification,
            "region": str(self.region),
        }

        if self.password:
            request["password"] = self.password

        if self.id:
            request["id"] = self.id

        return request



from typing import List, Dict, Callable, Optional
from http import HTTPStatus
from urllib.request import urlopen
import json
import requests
from functools import wraps

from jose import jwt


class OIDCAuthError(Exception):
    def __init__(self, message: str):
        self.message = message


class InsufficientPermissionError(Exception):
    def __init__(self, message: str):
        self.message = message


class OpenIDConnectAuth:
    def __init__(
        self, issuer_url: str, audience: str, algorithms: Optional[List[str]] = None,
    ):
        self._issuer_url = issuer_url
        self._audience = audience
        self._algorithms = algorithms or ["RS256"]

    def authorize(self, access_token: str, required_permission: str = None) -> str:
        self.validate_access_token(access_token)
        if required_permission and not self._has_permission(
            access_token, required_permission
        ):
            raise InsufficientPermissionError("Insufficient user permission")
        return self._get_user_email(access_token)

    def validate_access_token(self, access_token: str):
        unverified_header = jwt.get_unverified_header(access_token)
        jw_key_set = self._get_jw_key_set()
        rsa_key = self._find_rsa_key(unverified_header["kid"], jw_key_set["keys"])
        jwt.decode(
            access_token,
            rsa_key,
            algorithms=self._algorithms,
            audience=self._audience,
            issuer=self._issuer_url + "/",
        )

    def _get_user_email(self, access_token: str) -> str:
        response = requests.get(
            f"{self._issuer_url}/userinfo",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        if response.status_code != HTTPStatus.OK:
            raise OIDCAuthError("Invalid access token")
        data = response.json()
        if "email" not in data:
            raise OIDCAuthError("Attribute `email` doesn't exist in userinfo.")
        return data["email"]

    def _get_jw_key_set(self) -> Dict:
        response = requests.get(f"{self._issuer_url}/.well-known/jwks.json")
        return response.json()

    @staticmethod
    def _find_rsa_key(key_id: str, all_keys: List[Dict]) -> Dict:
        for key in all_keys:
            if key["kid"] == key_id:
                return {
                    "kty": key["kty"],
                    "kid": key["kid"],
                    "use": key["use"],
                    "n": key["n"],
                    "e": key["e"],
                }

    @staticmethod
    def _has_permission(access_token: str, required_permission: str) -> bool:
        unverified_claims = jwt.get_unverified_claims(access_token)
        permissions = unverified_claims.get("permissions", [])
        if required_permission in permissions:
            return True
        wildcard_permissions = [p for p in permissions if p.endswith("*")]
        for permission in wildcard_permissions:
            if required_permission.startswith(permission.strip("*")):
                return True
        return False


def extract_access_token(authorization_header: str) -> str:
    """Extract Bearer access token from `Authorization` header."""
    if not authorization_header:
        raise OIDCAuthError("Authorization header is expected")
    try:
        bearer, token = authorization_header.split()
    except ValueError:
        raise OIDCAuthError("Invalid bearer token format in authorization header")
    if bearer.lower() != "bearer":
        raise OIDCAuthError("Invalid bearer token format in authorization header")

    return token



import setuptools

setuptools.setup(
    name="user-sdk",
    version="0.0.52",
    description="Interface for user auth/profile backend",
    url="https://github.com/shuttl-tech/user_sdk",
    author="Shuttl",
    author_email="paul.kuruvilla@shuttl.com",
    license="MIT",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["requests", "pyshuttlis", "shuttl-geo", "python-jose"],
    extras_require={
        "test": ["pytest", "pytest-runner", "pytest-cov", "pytest-pep8", "responses"]
    },
)



class ClientError(Exception):
    def __init__(self, error=None):
        self.error = error


class NoSuchUser(ClientError):
    pass


class NoSuchProfile(ClientError):
    pass


class NoRoleFound(ClientError):
    pass


class ProfileCreationError(ClientError):
    pass


class ProfileUpdateError(ClientError):
    pass


class UserAlreadyExists(ClientError):
    pass


class InvalidOTPError(ClientError):
    pass


class InvalidRoleRequest(ClientError):
    pass


class ProfileAlreadyExists(ClientError):
    pass


class SessionNotFound(ClientError):
    pass


class TypeNotSupported(ClientError):
    pass


class TooManyRequests(ClientError):
    pass


class ServerError(Exception):
    def __init__(self, error=None):
        self.error = error


class OTPSendFailure(ServerError):
    pass


class UserCreationFailed(ServerError):
    pass


class UserAuthenticationError(ServerError):
    pass



import logging
import json


logger = logging.getLogger(__name__)


def log(
    message: str,
    request_url: str,
    status_code: str,
    response: str,
    request_body: dict = None,
):

    dict = {
        "request_url": request_url,
        "status_code": status_code,
        "response": response,
    }
    if request_body:
        dict["request_body"] = str(request_body)
    logger.debug(message, extra=dict)
